VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cJpeg"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
Option Base 0

'Class Name:   cJpeg.cls  "JPEG Encoder Class"
'Author:       John Korejwa  <korejwa@tiac.net>
'Version:      0.9 beta  [26 / November / 2003]
'
'
'Legal:
'        This class is intended for and was uploaded to www.planetsourcecode.com
'
'        This product includes JPEG compression code developed by John Korejwa.  <korejwa@tiac.net>
'        Source code, written in Visual Basic, is freely available for non-commercial,
'        non-profit use at www.planetsourcecode.com.
'
'
'Credits:
'        Special thanks to Barry G., a government research scientist who took an interest in my
'        steganography software and research in late 1999.  I never met Barry in person, but he
'        was kind enough to buy and mail me a book with the ISO DIS 10918-1 JPEG standard.
'
'
'Description:  This class contains code for compressing pictures, sampled via hDC, into
'              baseline .JPG files.  Please report any errors or unusual behavior to the email
'              address above.
'
'Dependencies: None
'

'JPEG Marker Constants                (Note: VB compiler does not compile unused constants)
'Non-Differential Huffman Coding
Private Const SOF0    As Long = &HC0& 'Baseline DCT
Private Const SOF1    As Long = &HC1& 'Extended sequential DCT
Private Const SOF2    As Long = &HC2& 'Progressive DCT
Private Const SOF3    As Long = &HC3& 'Spatial (sequential) lossless
'Differential Huffman coding
Private Const SOF5    As Long = &HC5& 'Differential sequential DCT
Private Const SOF6    As Long = &HC6& 'Differential progressive DCT
Private Const SOF7    As Long = &HC7& 'Differential spatial
'Non-Differential arithmetic coding
Private Const JPG     As Long = &HC8& 'Reserved for JPEG extentions
Private Const SOF9    As Long = &HC9& 'Extended sequential DCT
Private Const SOF10   As Long = &HCA& 'Progressive DCT
Private Const SOF11   As Long = &HCB& 'Spatial (sequential) lossless
'Differential arithmetic coding
Private Const SOF13   As Long = &HCD& 'Differential sequential DCT
Private Const SOF14   As Long = &HCE& 'Differential progressive DCT
Private Const SOF15   As Long = &HCF& 'Differential Spatial
'Other Markers
Private Const DHT     As Long = &HC4& 'Define Huffman tables
Private Const DAC     As Long = &HCC& 'Define arithmetic coding conditioning(s)
Private Const RSTm    As Long = &HD0& 'Restart with modulo 8 count "m"
Private Const RSTm2   As Long = &HD7& 'to 'Restart with modulo 8 count "m"
Private Const SOI     As Long = &HD8& 'Start of image
Private Const EOI     As Long = &HD9& 'End of image
Private Const SOS     As Long = &HDA& 'Start of scan
Private Const DQT     As Long = &HDB& 'Define quantization table(s)
Private Const DNL     As Long = &HDC& 'Define number of lines
Private Const DRI     As Long = &HDD& 'Define restart interval
Private Const DHP     As Long = &HDE& 'Define hierarchical progression
Private Const EXP     As Long = &HDF& 'Expand reference components
Private Const APP0    As Long = &HE0& 'Reserved for application segments
Private Const APPF    As Long = &HEF& '  to Reserved for application segments
Private Const JPGn    As Long = &HF0& 'Reserved for JPEG Extentions
Private Const JPGn2   As Long = &HFD& '  to Reserved for JPEG Extentions
Private Const COM     As Long = &HFE& 'Comment
Private Const RESm    As Long = &H2&  'Reserved
Private Const RESm2   As Long = &HBF& '  to Reserved
Private Const TEM     As Long = &H1&  'For temporary use in arithmetic coding

'Consider these arrays of constants.
'They are initialized with the class and do not change.
Private QLumin(63)    As Integer 'Standard Luminance   Quantum (for 50% quality)
Private QChrom(63)    As Integer 'Standard Chrominance Quantum (for 50% quality)
Private FDCTScale(7)  As Double  'Constants for scaling FDCT Coefficients
Private IDCTScale(7)  As Double  'Constants for scaling IDCT Coefficients
Private ZigZag(7, 7)  As Long    'Zig Zag order of 8X8 block of samples

'API constants
Private Const BLACKONWHITE    As Long = 1 'nStretchMode constants for
Private Const COLORONCOLOR    As Long = 3 '  SetStretchBltMode() API function
Private Const HALFTONE        As Long = 4 'HALFTONE not supported in Win 95, 98, ME

Private Const BI_RGB          As Long = 0
Private Const DIB_RGB_COLORS  As Long = 0

'Variable types needed for DIBSections.

Private Type SAFEARRAYBOUND
   cElements         As Long
   lLbound           As Long
End Type

Private Type SAFEARRAY2D
   cDims             As Integer
   fFeatures         As Integer
   cbElements        As Long
   cLocks            As Long
   pvData            As Long
   Bounds(0 To 1)    As SAFEARRAYBOUND
End Type

Private Type RGBQUAD
   rgbBlue           As Byte
   rgbGreen          As Byte
   rgbRed            As Byte
   rgbReserved       As Byte
End Type

Private Type BITMAPINFOHEADER
   biSize            As Long
   biWidth           As Long
   biHeight          As Long
   biPlanes          As Integer
   biBitCount        As Integer
   biCompression     As Long
   biSizeImage       As Long
   biXPelsPerMeter   As Long
   biYPelsPerMeter   As Long
   biClrUsed         As Long
   biClrImportant    As Long
End Type

Private Type BITMAPINFO
   bmiHeader         As BITMAPINFOHEADER
   bmiColors         As RGBQUAD
End Type

'API needed for creating DIBSections for sampling and pixel access.
Private Declare Function CreateCompatibleDC Lib "gdi32" (ByVal hDC As Long) As Long
Private Declare Function CreateDIBSection2 Lib "gdi32" _
      Alias "CreateDIBSection" ( _
      ByVal hDC As Long, _
      pBitmapInfo As BITMAPINFO, _
      ByVal un As Long, _
      lplpVoid As Long, _
      ByVal handle As Long, _
      ByVal dw As Long) As Long   'lplpVoid changed to ByRef
Private Declare Function BitBlt Lib "gdi32" ( _
      ByVal hDestDC As Long, _
      ByVal x As Long, _
      ByVal y As Long, _
      ByVal nWidth As Long, _
      ByVal nHeight As Long, _
      ByVal hSrcDC As Long, _
      ByVal xSrc As Long, _
      ByVal ySrc As Long, _
      ByVal dwRop As Long) As Long
Private Declare Function SelectObject Lib "gdi32" (ByVal hDC As Long, ByVal hObject As Long) As Long
Private Declare Function DeleteObject Lib "gdi32" (ByVal hObject As Long) As Long
Private Declare Function DeleteDC Lib "gdi32" (ByVal hDC As Long) As Long
Private Declare Function VarPtrArray Lib "msvbvm60.dll" Alias "VarPtr" (Ptr() As Any) As Long
Private Declare Function SetStretchBltMode Lib "gdi32" ( _
      ByVal hDC As Long, _
      ByVal nStretchMode As Long) As Long
Private Declare Function StretchBlt Lib "gdi32" ( _
      ByVal hDC As Long, _
      ByVal x As Long, _
      ByVal y As Long, _
      ByVal nWidth As Long, _
      ByVal nHeight As Long, _
      ByVal hSrcDC As Long, _
      ByVal xSrc As Long, _
      ByVal ySrc As Long, _
      ByVal nSrcWidth As Long, _
      ByVal nSrcHeight As Long, _
      ByVal dwRop As Long) As Long
Private Declare Sub CopyMemory Lib "kernel32" _
      Alias "RtlMoveMemory" ( _
      lpvDest As Any, _
      lpvSource As Any, _
      ByVal cbCopy As Long)

'Custom variable types used for this JPEG encoding implementation

Private Type QUANTIZATIONTABLE
   Qk(63)            As Integer 'Quantization Values
   FScale(63)        As Single  'Multiplication values to scale and Quantize   FDCT output
   IScale(63)        As Single  'Multiplication values to scale and DeQuantize IDCT input
End Type

Private Type HUFFMANTABLE
   BITS(15)          As Byte    'Number of huffman codes of length i+1
   HUFFVAL(255)      As Byte    'Huffman symbol values
   EHUFSI(255)       As Long    'Huffman code size for symbol i
   EHUFCO(255)       As Long    'Huffman code      for symbol i
   MINCODE(15)       As Long    '
   MAXCODE(15)       As Long    'Largest code value for length i+1
End Type

Private Type COMPONENT
   Ci                As Long    'Component ID                       [0-255]
   Hi                As Long    'Horizontal Sampling Factor         [1-4]
   Vi                As Long    'Vertical   Sampling Factor         [1-4]
   Tqi               As Long    'Quantization Table Select          [0-3]
   data()            As Integer 'DCT Coefficients
End Type

Private PP            As Long    'Sample Precision [8, 12]
Private YY            As Long    'Number of lines             [Image Height]
Private XX            As Long    'Number of samples per line  [Image Width]
Private Nf            As Long    'Number of components in Frame

Private HMax          As Long    'Maximum horizontal sampling frequency
Private VMax          As Long    'Maximum vertical   sampling frequency

Private m_Data()      As Byte    'JPEG File Data
Private m_Chr         As Long    'Current Character in m_Data
Private m_Ptr         As Long    'Byte index in m_Data
Private m_Bit         As Long    'Bit  index in m_Chr

Private m_Block(7, 7) As Single  'Buffer for calculating DCT

Private QTable(3)     As QUANTIZATIONTABLE  '4 Quantization Tables
Private HuffDC(3)     As HUFFMANTABLE       '4 DC Huffman Tables
Private HuffAC(3)     As HUFFMANTABLE       '4 AC Huffman Tables
Private Comp()        As COMPONENT          'Scan Components

Private m_Quality     As Long
Private m_Comment     As String

'========================================================================================
'              D I S C R E T E   C O S I N E   T R A N S F O R M A T I O N
'========================================================================================

Private Sub Class_Initialize()

'========================================================================================
'                         C L A S S   I N I T I A L I Z A T I O N
'========================================================================================

  Dim i As Long
  Dim j As Long
  Dim dX As Long
  Dim zz As Long

   i = 0                   'Initialize the ZigZag() array, which maps out the
   j = 0                   '  zig-zag sequence of quantized DCT coefficients
   dX = 1                  '  in approximately low to high spatial frequencies

   For zz = 0 To 63
      ZigZag(i, j) = zz
      i = i + dX
      j = j - dX
      If i > 7 Then              '  0   1   5   6  14  15  27  28
         i = 7                  '  2   4   7  13  16  26  29  42
         j = j + 2              '  3   8  12  17  25  30  41  43
         dX = -1                '  9  11  18  24  31  40  44  53
       ElseIf j > 7 Then          ' 10  19  23  32  39  45  52  54
         j = 7                  ' 20  22  33  38  46  51  55  60
         i = i + 2              ' 21  34  37  47  50  56  59  61
         dX = 1                 ' 35  36  48  49  57  58  62  63
       ElseIf i < 0 Then
         i = 0 'check (j>7) first
         dX = 1
       ElseIf j < 0 Then
         j = 0
         dX = -1
      End If

   Next zz

   'Luminance Quantization table for Quality = 50
   QLumin(0) = 16:   QLumin(1) = 11:   QLumin(2) = 12:   QLumin(3) = 14
   QLumin(4) = 12:   QLumin(5) = 10:   QLumin(6) = 16:   QLumin(7) = 14
   QLumin(8) = 13:   QLumin(9) = 14:   QLumin(10) = 18:  QLumin(11) = 17
   QLumin(12) = 16:  QLumin(13) = 19:  QLumin(14) = 24:  QLumin(15) = 40
   QLumin(16) = 26:  QLumin(17) = 24:  QLumin(18) = 22:  QLumin(19) = 22
   QLumin(20) = 24:  QLumin(21) = 49:  QLumin(22) = 35:  QLumin(23) = 37
   QLumin(24) = 29:  QLumin(25) = 40:  QLumin(26) = 58:  QLumin(27) = 51
   QLumin(28) = 61:  QLumin(29) = 60:  QLumin(30) = 57:  QLumin(31) = 51
   QLumin(32) = 56:  QLumin(33) = 55:  QLumin(34) = 64:  QLumin(35) = 72
   QLumin(36) = 92:  QLumin(37) = 78:  QLumin(38) = 64:  QLumin(39) = 68
   QLumin(40) = 87:  QLumin(41) = 69:  QLumin(42) = 55:  QLumin(43) = 56
   QLumin(44) = 80:  QLumin(45) = 109: QLumin(46) = 81:  QLumin(47) = 87
   QLumin(48) = 95:  QLumin(49) = 98:  QLumin(50) = 103: QLumin(51) = 104
   QLumin(52) = 103: QLumin(53) = 62:  QLumin(54) = 77:  QLumin(55) = 113
   QLumin(56) = 121: QLumin(57) = 112: QLumin(58) = 100: QLumin(59) = 120
   QLumin(60) = 92:  QLumin(61) = 101: QLumin(62) = 103: QLumin(63) = 99

   'Chrominance Quantization table for Quality = 50
   QChrom(0) = 17:   QChrom(1) = 18:   QChrom(2) = 18:   QChrom(3) = 24
   QChrom(4) = 21:   QChrom(5) = 24:   QChrom(6) = 47:   QChrom(7) = 26
   QChrom(8) = 26:   QChrom(9) = 47:   QChrom(10) = 99:  QChrom(11) = 66
   QChrom(12) = 56:  QChrom(13) = 66:  QChrom(14) = 99:  QChrom(15) = 99
   QChrom(16) = 99:  QChrom(17) = 99:  QChrom(18) = 99:  QChrom(19) = 99
   QChrom(20) = 99:  QChrom(21) = 99:  QChrom(22) = 99:  QChrom(23) = 99
   QChrom(24) = 99:  QChrom(25) = 99:  QChrom(26) = 99:  QChrom(27) = 99
   QChrom(28) = 99:  QChrom(29) = 99:  QChrom(30) = 99:  QChrom(31) = 99
   QChrom(32) = 99:  QChrom(33) = 99:  QChrom(34) = 99:  QChrom(35) = 99
   QChrom(36) = 99:  QChrom(37) = 99:  QChrom(38) = 99:  QChrom(39) = 99
   QChrom(40) = 99:  QChrom(41) = 99:  QChrom(42) = 99:  QChrom(43) = 99
   QChrom(44) = 99:  QChrom(45) = 99:  QChrom(46) = 99:  QChrom(47) = 99
   QChrom(48) = 99:  QChrom(49) = 99:  QChrom(50) = 99:  QChrom(51) = 99
   QChrom(52) = 99:  QChrom(53) = 99:  QChrom(54) = 99:  QChrom(55) = 99
   QChrom(56) = 99:  QChrom(57) = 99:  QChrom(58) = 99:  QChrom(59) = 99
   QChrom(60) = 99:  QChrom(61) = 99:  QChrom(62) = 99:  QChrom(63) = 99

   FDCTScale(0) = 0.353553390593273     '0.25 / Cos(4 / 16 * PI)
   FDCTScale(1) = 0.25489778955208      '0.25 / Cos(1 / 16 * PI)
   FDCTScale(2) = 0.270598050073098     '0.25 / Cos(2 / 16 * PI)
   FDCTScale(3) = 0.300672443467523     '0.25 / Cos(3 / 16 * PI)
   FDCTScale(4) = 0.353553390593273     '0.25 / Cos(4 / 16 * PI)
   FDCTScale(5) = 0.449988111568207     '0.25 / Cos(5 / 16 * PI)
   FDCTScale(6) = 0.653281482438186     '0.25 / Cos(6 / 16 * PI)
   FDCTScale(7) = 1.28145772387074      '0.25 / Cos(7 / 16 * PI)

   SetSamplingFrequencies 2, 2, 1, 1, 1, 1
   Quality = 75

End Sub

Private Sub CollectStatisticsAC(data() As Integer, freqac() As Long)

'========================================================================================
'                      C O L L E C T I N G   S T A T I S T I C S
'========================================================================================
'These procedures collect statistics of run-length and size categories of DCT coefficients
'so optimized Huffman tables can be generated to compress them.

  Dim code As Long
  Dim n    As Long 'Number of coefficients in data()
  Dim p    As Long 'Index for current data() coefficient
  Dim p2   As Long
  Dim r    As Long 'Run length of zeros
  Dim rs   As Long 'Run-length/Size-category Symbol

   n = UBound(data) + 1
   p = 0

   While p <> n
      p = p + 1     'Skip DC coefficient
      p2 = p + 63   '63 AC coefficients

      r = 0

      While p <> p2

         If data(p) = 0 Then
            r = r + 1
          Else

            While r > 15
               freqac(240) = freqac(240) + 1  'RUN16 Symbol
               r = r - 16
            Wend

            code = data(p)
            If code < 0 Then 'rs = number of bits needed for code
               rs = Int((Log(-code) * 1.442695040889)) + 1   '1/log(2)  (+ error correction)
             ElseIf code > 0 Then
               rs = Int((Log(code) * 1.442695040889)) + 1    '1/log(2)  (+ error correction)
             Else
               rs = 0
            End If

            rs = (r * 16) Or rs
            freqac(rs) = freqac(rs) + 1        'Run-length/Size-category Symbol
            r = 0
         End If

         p = p + 1
      Wend

      If r <> 0 Then freqac(0) = freqac(0) + 1   'EOB Symbol
   Wend

End Sub

Private Sub CollectStatisticsDCInterleaved(data() As Integer, _
                                           freqdc() As Long, _
                                           Hi As Long, _
                                           Vi As Long)

  Dim p()       As Long     'Index to .data in component f for scanline g
  Dim f         As Long      'Index counter  (component)
  Dim g         As Long      'Index counter  (sampling factor, vertical)
  Dim h         As Long      'Index counter  (sampling factor, horizontal)
  Dim i         As Long      'Index counter  (MCU horizontal)
  Dim j         As Long      'Index counter  (MCU vertical)
  Dim n         As Long      'Number of coefficients in data()
  Dim s         As Long      'Size category for Diff
  Dim Diff      As Long      'DC Difference
  Dim Pred      As Long      'DC Predictor
  Dim pLF       As Long      'Line Feed for p in data
  Dim MCUr      As Long      'Number of complete 8X8 blocks in rightmost MCU
  Dim MCUx      As Long      'Number of MCUs per scanline
  Dim MCUy      As Long      'Number of MCU scanlines

   n = UBound(data) + 1
   ReDim p(Vi - 1)

   MCUx = (XX + 8 * HMax - 1) \ (8 * HMax)
   MCUy = (YY + 8 * VMax - 1) \ (8 * VMax)

   h = (-Int(-XX * Hi / HMax) + 7) \ 8
   'h = h \ 8                            'Width of scanline in data (MCUs)

   For g = 0 To Vi - 1                  'Initialize .data pointers
      p(g) = 64 * h * g
   Next g

   pLF = 64 * h * (Vi - 1)              'Initialize .data pointer advancer

   MCUr = (h Mod Hi)                    'Number of complete 8X8 Blocks in rightmost MCU
   If MCUr = 0 Then MCUr = Hi

   For j = 1 To MCUy - 1

      'MCUs across a scanline

      For i = 1 To MCUx - 1
         For g = 1 To Vi
            For h = 1 To Hi

               Diff = data(p(g - 1)) - Pred
               Pred = data(p(g - 1))
               p(g - 1) = p(g - 1) + 64
               If Diff < 0 Then 's = number of bits needed for Diff
                  s = Int((Log(-Diff) * 1.442695040889)) + 1   '1/log(2)  (+ error correction)
                ElseIf Diff > 0 Then
                  s = Int((Log(Diff) * 1.442695040889)) + 1    '1/log(2) + (error correction)
                Else
                  s = 0
               End If

               freqdc(s) = freqdc(s) + 1

            Next h
         Next g
      Next i

      'Rightmost MCU

      For g = 1 To Vi
         For h = 1 To Hi
            If h > MCUr Then  'Pad with dummy block
               s = 0
             Else
               Diff = data(p(g - 1)) - Pred
               Pred = data(p(g - 1))
               p(g - 1) = p(g - 1) + 64

               If Diff < 0 Then
                  s = Int((Log(-Diff) * 1.442695040889)) + 1
                ElseIf Diff > 0 Then
                  s = Int((Log(Diff) * 1.442695040889)) + 1
                Else
                  s = 0
               End If

            End If
            freqdc(s) = freqdc(s) + 1
         Next h

      Next g

      'Advance data pointers

      For g = 0 To Vi - 1
         p(g) = p(g) + pLF
      Next g

   Next j

   'Bottommost MCU Scanline

   For i = 1 To MCUx
      For g = 1 To Vi
         For h = 1 To Hi
            If p(g - 1) >= n Or (i = MCUx And h > MCUr) Then 'Pad with dummy block
               s = 0
             Else
               Diff = data(p(g - 1)) - Pred
               Pred = data(p(g - 1))
               p(g - 1) = p(g - 1) + 64

               If Diff < 0 Then
                  s = Int((Log(-Diff) * 1.442695040889)) + 1
                ElseIf Diff > 0 Then
                  s = Int((Log(Diff) * 1.442695040889)) + 1
                Else
                  s = 0
               End If

            End If
            freqdc(s) = freqdc(s) + 1
         Next h

      Next g
   Next i

End Sub

Private Sub CollectStatisticsDCNonInterleaved(data() As Integer, freqdc() As Long)

  Dim Diff  As Long     'DC Difference
  Dim Pred  As Long     'DC Predictor
  Dim n     As Long     'Number of coefficients in data()
  Dim p     As Long     'Index for current data() coefficient
  Dim s     As Long     'Size category for Diff

   n = UBound(data) + 1
   p = 0
   Pred = 0

   While p <> n
      Diff = data(p) - Pred
      Pred = data(p)

      If Diff < 0 Then 's = number of bits needed for Diff
         s = Int((Log(-Diff) * 1.442695040889)) + 1   '1/log(2)  (+ error correction)
       ElseIf Diff > 0 Then
         s = Int((Log(Diff) * 1.442695040889)) + 1    '1/log(2) + (error correction)
       Else
         s = 0
      End If

      freqdc(s) = freqdc(s) + 1
      p = p + 64
   Wend

End Sub

Public Property Get Comment() As String

   Comment = m_Comment

End Property

Public Property Let Comment(Value As String)

   'Assigning a value to this property will add the text Comment to the JPEG file.
   If Len(Value) > 65535 Then Err.Raise 1, , "Illegal Comment Length"
   m_Comment = Value

End Property

Private Sub EncodeCoefficients(data() As Integer, p As Long, Pred As Long, Td As Long, Ta As Long)

   'Use Huffman tables to compress a block of 64 quantized DCT coefficients to the local
   'm_Data() byte array.  The coefficients are input in the data() array starting at index p.
   'Pred is the predictor for the DC coefficient.  Td and Ta are indexes to the local DC and AC
   'Huffman Tables to use.
  Dim r     As Long
  Dim rs    As Long
  Dim si    As Long
  Dim code  As Long
  Dim p2    As Long

   p2 = p + 64

   code = data(p) - Pred
   Pred = data(p)
   p = p + 1

   si = 1
   rs = 0

   If code < 0 Then

      Do While si <= -code
         si = si * 2
         rs = rs + 1
      Loop

      code = code - 1
    Else

      Do While si <= code
         si = si * 2
         rs = rs + 1
      Loop

   End If
   si = si \ 2
   WriteBits HuffDC(Td).EHUFSI(rs), HuffDC(Td).EHUFCO(rs) 'append symbol for size category
   WriteBits si, code                                     'append diff

   With HuffAC(Ta)
      r = 0

      Do

         If data(p) = 0 Then
            r = r + 1
          Else

            While r > 15
               WriteBits .EHUFSI(240), .EHUFCO(240) 'append RUN16 (a run of 16 zeros)
               r = r - 16
            Wend

            code = data(p)
            rs = r * 16
            si = 1

            If code < 0 Then

               Do While si <= -code
                  si = si * 2
                  rs = rs + 1
               Loop

               code = code - 1
             Else

               Do While si <= code
                  si = si * 2
                  rs = rs + 1
               Loop

            End If
            si = si \ 2
            WriteBits .EHUFSI(rs), .EHUFCO(rs) 'append run length, size category
            WriteBits si, code                 'append AC value
            r = 0
         End If

         p = p + 1
      Loop While p < p2 'should be equal on exit

      If r <> 0 Then WriteBits .EHUFSI(0), .EHUFCO(0) 'append EOB (end of block)
   End With

End Sub

Private Sub ExpandDQT(Tqi As Long)

'========================================================================================
'                                Q U A N T I Z A T I O N
'========================================================================================

  Dim i          As Long
  Dim j          As Long
  Dim k          As Byte
  Dim maxvalue   As Long

   With QTable(Tqi)

      If PP = 12 Then
         maxvalue = 65535
       Else
         maxvalue = 255
      End If

      For i = 0 To 7
         For j = 0 To 7
            k = ZigZag(i, j)
            If .Qk(k) < 1 Or .Qk(k) > maxvalue Then Err.Raise 1, , "Bad Quantization Table"
            .FScale(k) = FDCTScale(i) * FDCTScale(j) / CDbl(.Qk(k))
         Next j

      Next i
   End With

End Sub

Private Sub ExpandHuffman(TheHuff As HUFFMANTABLE, Optional MaxSymbol As Long = 255)

   'Given a HUFFMANTABLE with valid BITS and HUFFVAL, generate tables for
   'EHUFCO, EHUFSI, MAXCODE, and MINCODE so the table may be used for compression
   'and/or decompression.  In JPEG, MaxSymbol is 255 for an AC Huffman Table.  For
   'DC Tables, MaxSymbol is 11 for PP=8 bit precission, or 15 for PP=12 bit precission.
  Dim i          As Long 'Index for BITS
  Dim j          As Long 'Index for HUFFVAL
  Dim k          As Long 'Index for last HUFFVAL of length (i+1)
  Dim si         As Long 'Huffman code size  ( =2^i )
  Dim code       As Long 'Huffman code
  Dim symbol     As Long 'Huffman symbol

   With TheHuff

      For i = 0 To 255
         .EHUFSI(i) = 0      'Clear existing values so we can
         .EHUFCO(i) = -1     'check for duplicate huffman symbols
      Next i

      j = 0
      si = 1
      code = 0

      For i = 0 To 15
         k = j + .BITS(i)
         If k > 256 Then Err.Raise 1, , "Bad Huffman Table" 'more than 256 symbols
         If j = k Then 'no codes of length i+1
            .MINCODE(i) = j - code
            .MAXCODE(i) = -1
          Else
            .MINCODE(i) = j - code

            While j < k
               symbol = .HUFFVAL(j)  'read symbol, make sure it's valid
               If symbol > MaxSymbol Then Err.Raise 1, , "Bad Huffman Table"   'invalid symbol
               If .EHUFCO(symbol) >= 0 Then Err.Raise 1, , "Bad Huffman Table" 'duplicate symbol
               .EHUFSI(symbol) = si    'assign code for symbol
               .EHUFCO(symbol) = code
               code = code + 1
               j = j + 1
            Wend

            .MAXCODE(i) = code - 1
         End If

         si = si * 2
         If code >= si Then Err.Raise 1, , "Bad Huffman Table" 'code does not fit into available
                                                               '   bits
         code = code * 2
      Next i

      If j = 0 Then Err.Raise 1, , "Bad Huffman Table" 'No huffman symbols???
   End With

End Sub

Private Sub FDCT()

  Static t0   As Single 'Given an 8X8 block of discretely sampled values [m_Block(0-7, 0-7)],
  Static t1   As Single 'replace them with their (scaled) Forward Discrete Cosine Transformation
                        '   values.
  Static t2   As Single '80 (+64) multiplications and 464 additions are needed.
  Static t3   As Single 'Values are scaled on output, meaning that each of the 64 elements must be
  Static t4   As Single 'multiplied by constants for a final FDCT.  These final constants are
                        '   combined
  Static t5   As Single 'with Quantization constants, so a final 64 multiplications combine the
  Static t6   As Single 'completion of the FDCT and Quantization in one step.
  Static t7   As Single
  Static t8   As Single
  Static i    As Long

   For i = 0 To 7                  'Process 1D FDCT on each row
      t0 = m_Block(i, 0) + m_Block(i, 7)
      t1 = m_Block(i, 0) - m_Block(i, 7)
      t2 = m_Block(i, 1) + m_Block(i, 6)
      t3 = m_Block(i, 1) - m_Block(i, 6)
      t4 = m_Block(i, 2) + m_Block(i, 5)
      t5 = m_Block(i, 2) - m_Block(i, 5)
      t6 = m_Block(i, 3) + m_Block(i, 4)
      t7 = m_Block(i, 3) - m_Block(i, 4)

      t7 = t7 + t5
      t8 = t0 - t6
      t6 = t6 + t0
      t0 = t2 + t4
      t2 = (t2 - t4 + t8) * 0.707106781186548   'Cos(2# * PI / 8#)
      t4 = t1 + t3
      t3 = (t3 + t5) * 0.707106781186548        'Cos(2# * PI / 8#)
      t5 = (t4 - t7) * 0.382683432365091        'Cos(3# * PI / 8#)
      t7 = t7 * 0.541196100146196 - t5          'Cos(PI / 8#) - Cos(3# * PI / 8#)
      t4 = t4 * 1.30656296487638 - t5           'Cos(PI / 8#) + Cos(3# * PI / 8#)
      t5 = t1 + t3
      t1 = t1 - t3

      m_Block(i, 0) = t6 + t0
      m_Block(i, 4) = t6 - t0
      m_Block(i, 1) = t5 + t4
      m_Block(i, 7) = t5 - t4
      m_Block(i, 2) = t8 + t2
      m_Block(i, 6) = t8 - t2
      m_Block(i, 5) = t1 + t7
      m_Block(i, 3) = t1 - t7
   Next i

   For i = 0 To 7                   'Process 1D FDCT on each column
      t0 = m_Block(0, i) + m_Block(7, i)
      t1 = m_Block(0, i) - m_Block(7, i)
      t2 = m_Block(1, i) + m_Block(6, i)
      t3 = m_Block(1, i) - m_Block(6, i)
      t4 = m_Block(2, i) + m_Block(5, i)
      t5 = m_Block(2, i) - m_Block(5, i)
      t6 = m_Block(3, i) + m_Block(4, i)
      t7 = m_Block(3, i) - m_Block(4, i)

      t7 = t7 + t5
      t8 = t0 - t6
      t6 = t6 + t0
      t0 = t2 + t4
      t2 = (t2 - t4 + t8) * 0.707106781186548   'Cos(2# * PI / 8#)
      t4 = t1 + t3
      t3 = (t3 + t5) * 0.707106781186548        'Cos(2# * PI / 8#)
      t5 = (t4 - t7) * 0.382683432365091        'Cos(3# * PI / 8#)
      t7 = t7 * 0.541196100146196 - t5          'Cos(PI / 8#) - Cos(3# * PI / 8#)
      t4 = t4 * 1.30656296487638 - t5           'Cos(PI / 8#) + Cos(3# * PI / 8#)
      t5 = t1 + t3
      t1 = t1 - t3

      m_Block(0, i) = t6 + t0
      m_Block(4, i) = t6 - t0
      m_Block(1, i) = t5 + t4
      m_Block(7, i) = t5 - t4
      m_Block(2, i) = t8 + t2
      m_Block(6, i) = t8 - t2
      m_Block(5, i) = t1 + t7
      m_Block(3, i) = t1 - t7
   Next i

End Sub

Private Sub InsertCOM(TheComment As String)

  Dim i As Long
  Dim Lx As Long

   Lx = Len(TheComment) + 2

   If Lx > 2 Then
      m_Data(m_Ptr) = 255               'COM marker
      m_Data(m_Ptr + 1) = COM
      m_Data(m_Ptr + 2) = Lx \ 256      'COM marker segment length
      m_Data(m_Ptr + 3) = Lx And 255
      m_Ptr = m_Ptr + 4

      For i = 1 To Len(TheComment)      'Comment text
         m_Data(m_Ptr) = Asc(Mid$(TheComment, i, 1))
         m_Ptr = m_Ptr + 1
      Next i

   End If

End Sub

Private Sub InsertDHT(ByVal MarkerPos As Long, HIndex As Long, IsAC As Boolean)

  Dim i As Long 'Call with MarkerPos = m_Ptr to insert a single table with its own DHT marker
  Dim j As Long 'Call multiple times with the same MarkerPos to include

   'multiple tables under the same DHT marker
   If m_Ptr < MarkerPos + 4 Then 'Insert Marker
      m_Ptr = MarkerPos + 4
      m_Data(m_Ptr - 4) = 255
      m_Data(m_Ptr - 3) = DHT
   End If

   If IsAC Then

      With HuffAC(HIndex)
         m_Data(m_Ptr) = HIndex Or 16
         m_Ptr = m_Ptr + 1
         j = 0

         For i = 0 To 15
            m_Data(m_Ptr) = .BITS(i)
            m_Ptr = m_Ptr + 1
            j = j + .BITS(i)
         Next i

         For i = 0 To j - 1
            m_Data(m_Ptr) = .HUFFVAL(i)
            m_Ptr = m_Ptr + 1
         Next i

      End With
    Else

      With HuffDC(HIndex)
         m_Data(m_Ptr) = HIndex
         m_Ptr = m_Ptr + 1
         j = 0

         For i = 0 To 15
            m_Data(m_Ptr) = .BITS(i)
            m_Ptr = m_Ptr + 1
            j = j + .BITS(i)
         Next i

         For i = 0 To j - 1
            m_Data(m_Ptr) = .HUFFVAL(i)
            m_Ptr = m_Ptr + 1
         Next i

      End With
   End If

   m_Data(MarkerPos + 2) = (m_Ptr - MarkerPos - 2) \ 256& 'Insert Marker segment length
   m_Data(MarkerPos + 3) = (m_Ptr - MarkerPos - 2) And 255&

End Sub

Private Sub InsertDQT(ByVal MarkerPos As Long, Tqi As Long)

  Dim i As Long 'Call with MarkerPos = m_Ptr to insert a single table with its own DQT marker

   'Call multiple times with the same MarkerPos to include
   'multiple tables under the same DQT marker

   If m_Ptr < MarkerPos + 4 Then 'Insert Marker
      m_Ptr = MarkerPos + 4
      m_Data(m_Ptr - 4) = 255
      m_Data(m_Ptr - 3) = DQT
   End If

   With QTable(Tqi)

      For i = 0 To 63
         If .Qk(i) > 255 Then Exit For
      Next i

      If i = 64 Then              '8 bit precision
         m_Data(m_Ptr) = Tqi
         m_Ptr = m_Ptr + 1

         For i = 0 To 63
            m_Data(m_Ptr) = .Qk(i)
            m_Ptr = m_Ptr + 1
         Next i

       Else                        '16 bit precision
         If PP <> 12 Then Err.Raise 1, , "Illegal precission in Quantization Table"
         m_Data(m_Ptr) = Tqi Or 16
         m_Ptr = m_Ptr + 1

         For i = 0 To 63
            m_Data(m_Ptr) = .Qk(i) \ 256
            m_Data(m_Ptr + 1) = .Qk(i) And 255
            m_Ptr = m_Ptr + 2
         Next i

      End If
   End With

   m_Data(MarkerPos + 2) = (m_Ptr - MarkerPos - 2) \ 256& 'Insert Marker segment length
   m_Data(MarkerPos + 3) = (m_Ptr - MarkerPos - 2) And 255&

End Sub

Private Sub InsertJFIF()

'================================================================================
'                         E M I T I N G   M A R K E R S
'================================================================================

   If m_Ptr + 17 > UBound(m_Data) Then Err.Raise 9 'Copymemory will write past bounds of m_Data()

   CopyMemory m_Data(m_Ptr + 0), &H1000E0FF, 4&    'APP0 Marker, Length(APP0)=16
   CopyMemory m_Data(m_Ptr + 4), &H4649464A, 4&    '"JFIF"
   CopyMemory m_Data(m_Ptr + 8), &H10100, 4&       '"/0", Version Major=1, Version Minor=1
   'Units=0  [0=pixel, 1=dpi, 2=dots/cm]
   CopyMemory m_Data(m_Ptr + 12), &H1000100, 4&    'Horizontal pixel density = 1 (dot per pixel)
   'Vertical   pixel density = 1 (dot per pixel)
   CopyMemory m_Data(m_Ptr + 16), &H0&, 2&         'Thumbnail horizontal pixel count = 0
   m_Ptr = m_Ptr + 18                              'Thumbnail vertical   pixel count = 0

End Sub

Private Sub InsertMarker(TheMarker As Long)

   m_Data(m_Ptr) = 255
   m_Data(m_Ptr + 1) = TheMarker
   m_Ptr = m_Ptr + 2

End Sub

Private Sub InsertSequentialScans(CompIndex() As Long, _
                                  Td() As Long, _
                                  Ta() As Long, _
                                  FirstIndex As Long, _
                                  SecondIndex As Long)

   'Insert scan components CompIndex(FirstIndex) to CompIndex(SecondIndex) sequentially in
   '   compliance
   'with JPEG rules.  Components are interleaved whenever possible to emit as few scans as possible.

  Dim f            As Long       'First Index
  Dim g            As Long       'Second Index
  Dim Nb           As Long       'Number of 8X8 blocks in MCU
  Const MaxNb      As Long = 10  'Max 8X8 blocks in MCU  (10 for JPEG compliance)
  Dim flag         As Boolean    'True when ready to insert scan(s)

   f = FirstIndex
   g = FirstIndex
   Nb = 0
   flag = False

   While f <= SecondIndex

      Nb = Nb + Comp(CompIndex(g)).Hi * Comp(CompIndex(g)).Vi
      g = g + 1

      If Nb > MaxNb Then
         flag = True
         If f <> g - 1 Then g = g - 1
       Else
         If (g - f) = 3 Or g > SecondIndex Then flag = True
      End If

      If flag Then
         If f = g - 1 Then
            InsertSOSNonInterleaved CompIndex(f), Td(f), Ta(f)
          Else
            InsertSOSInterleaved CompIndex, Td, Ta, f, g - 1
         End If

         Nb = 0
         f = g
         flag = False
      End If

   Wend

End Sub

Private Sub InsertSOF(SOFMarker As Long)

  Dim i   As Long 'Insert a Start Of Frame marker segment
  Dim Lx  As Long 'PP, YY, XX, Nf, and Ci,Hi,Vi,Tqi, must already be set

   Lx = 8 + 3 * Nf
   m_Data(m_Ptr) = 255                    'SOF
   m_Data(m_Ptr + 1) = SOFMarker And 255
   m_Data(m_Ptr + 2) = Lx \ 256           'Frame Header Length
   m_Data(m_Ptr + 3) = Lx And 255
   m_Data(m_Ptr + 4) = PP                 'Sample precision [8, 12]
   m_Data(m_Ptr + 5) = YY \ 256           'Number of Lines
   m_Data(m_Ptr + 6) = YY And 255
   m_Data(m_Ptr + 7) = XX \ 256           'Number of samples per line
   m_Data(m_Ptr + 8) = XX And 255
   m_Data(m_Ptr + 9) = Nf                 'Number of image components in frame
   m_Ptr = m_Ptr + 10

   For i = 0 To Nf - 1                      'For each component ...

      With Comp(i)
         m_Data(m_Ptr) = .Ci                  'Component identifier
         m_Data(m_Ptr + 1) = .Hi * 16 Or .Vi  'Horizontal/Vertical sampling factors
         m_Data(m_Ptr + 2) = .Tqi             'Quantization table selector
      End With

      m_Ptr = m_Ptr + 3
   Next i

End Sub

Private Sub InsertSOSInterleaved(CompIndex() As Long, _
                                 Td() As Long, _
                                 Ta() As Long, _
                                 FirstIndex As Long, _
                                 SecondIndex As Long)

   'Insert an SOS marker and scan data for an interleaved Sequential scan.

  Dim f         As Long      'Index counter  (component)
  Dim g         As Long      'Index counter  (sampling factor, vertical)
  Dim h         As Long      'Index counter  (sampling factor, horizontal)
  Dim i         As Long      'Index counter  (MCU horizontal)
  Dim j         As Long      'Index counter  (MCU vertical)
  Dim Lx        As Long      'Marker Segment Length
  Dim Ns        As Long      'Number of components in Scan [1-4]
  Dim MCUx      As Long      'Number of MCUs per scanline
  Dim MCUy      As Long      'Number of MCU scanlines

  Dim p()        As Long     'Index to .data in component f for scanline g
  Dim pLF()      As Long     'Line Feed for p in .data for component f
  Dim Pred()     As Long     'Predictor for DC coefficient in component f
  Dim MCUr()     As Long     'Number of complete 8X8 blocks in rightmost MCU for component f
  Dim Pad64(63)  As Integer  '8X8 padding block for completing MCUs

   Ns = SecondIndex - FirstIndex + 1
   Lx = 6 + 2 * Ns

   'Insert SOS Marker Segment
   m_Data(m_Ptr) = 255                          'SOS Marker
   m_Data(m_Ptr + 1) = SOS
   m_Data(m_Ptr + 2) = Lx \ 256                 'Marker Segment Length
   m_Data(m_Ptr + 3) = Lx And 255
   m_Data(m_Ptr + 4) = Ns                       'Ns     - Number of components in Scan [1-4]
   m_Ptr = m_Ptr + 5

   For i = FirstIndex To SecondIndex
      m_Data(m_Ptr) = Comp(CompIndex(i)).Ci   'Csj
      m_Data(m_Ptr + 1) = Td(i) * 16 Or Ta(i) 'Td, Ta
      m_Ptr = m_Ptr + 2
   Next i

   m_Data(m_Ptr) = 0                            'Ss     - Start of spectral selection
   m_Data(m_Ptr + 1) = 63                       'Se     - End of spectral selection
   m_Data(m_Ptr + 2) = 0                        'Ah, Al - Successive approximation bit high/low
   m_Ptr = m_Ptr + 3

   'Insert interleaved sequential entropy coded data
   ReDim p(FirstIndex To SecondIndex, VMax - 1)
   ReDim Pred(FirstIndex To SecondIndex)
   ReDim pLF(FirstIndex To SecondIndex)
   ReDim MCUr(FirstIndex To SecondIndex)

   MCUx = (XX + 8 * HMax - 1) \ (8 * HMax)
   MCUy = (YY + 8 * VMax - 1) \ (8 * VMax)

   For f = FirstIndex To SecondIndex

      With Comp(CompIndex(f))
         h = (-Int(-XX * .Hi / HMax) + 7) \ 8  'Width of scanline in .data (MCUs)

         For g = 0 To .Vi - 1                  'Initialize .data pointers
            p(f, g) = 64 * h * g
         Next g

         pLF(f) = 64 * h * (.Vi - 1)           'Initialize .data pointer advancer

         MCUr(f) = (h Mod .Hi)                 'Number of complete 8X8 Blocks in rightmost MCU
         If MCUr(f) = 0 Then MCUr(f) = .Hi
      End With

   Next f

   WriteBitsBegin

   For j = 1 To MCUy - 1

      'Encode MCUs across a scanline

      For i = 1 To MCUx - 1
         For f = FirstIndex To SecondIndex '0 To Ns - 1

            With Comp(CompIndex(f))

               For g = 1 To .Vi
                  For h = 1 To .Hi
                     EncodeCoefficients .data, p(f, g - 1), Pred(f), Td(f), Ta(f)
                  Next h

               Next g
            End With

         Next f
      Next i

      'Encode Rightmost MCU

      For f = FirstIndex To SecondIndex '0 To Ns - 1

         With Comp(CompIndex(f))

            For g = 1 To .Vi
               For h = 1 To .Hi
                  If h > MCUr(f) Then 'Pad with dummy block
                     Pad64(0) = Pred(f)
                     EncodeCoefficients Pad64, 0, Pred(f), Td(f), Ta(f)
                   Else
                     EncodeCoefficients .data, p(f, g - 1), Pred(f), Td(f), Ta(f)
                  End If

               Next h
            Next g
         End With

      Next f

      'Advance .data pointers

      For f = FirstIndex To SecondIndex
         For g = 0 To Comp(CompIndex(f)).Vi - 1
            p(f, g) = p(f, g) + pLF(f)
         Next g

      Next f
   Next j

   'Encode Bottommost MCU Scanline

   For i = 1 To MCUx
      For f = FirstIndex To SecondIndex

         With Comp(CompIndex(f))

            For g = 1 To .Vi
               For h = 1 To .Hi
                  If p(f, g - 1) > UBound(.data) Or (i = MCUx And h > MCUr(f)) Then 'Pad with dummy
                                                                                    '   block
                     Pad64(0) = Pred(f)
                     EncodeCoefficients Pad64, 0, Pred(f), Td(f), Ta(f)
                   Else
                     EncodeCoefficients .data, p(f, g - 1), Pred(f), Td(f), Ta(f)
                  End If

               Next h
            Next g
         End With

      Next f
   Next i

   WriteBitsEnd

End Sub

Private Sub InsertSOSNonInterleaved(CompIndex As Long, Td As Long, Ta As Long)

'================================================================================
'                           E M I T I N G   S C A N S
'================================================================================

   'Insert an SOS marker and scan data for a non-interleaved Sequential scan.
  Dim p         As Long     'Index to .data in component
  Dim n         As Long
  Dim Pred      As Long     'Predictor for DC coefficient

   'Insert SOS Marker Segment
   m_Data(m_Ptr) = 255                          'SOS Marker
   m_Data(m_Ptr + 1) = SOS
   m_Data(m_Ptr + 2) = 8 \ 256                  'Marker Segment Length
   m_Data(m_Ptr + 3) = 8 And 255
   m_Data(m_Ptr + 4) = 1                        'Ns     - Number of components in Scan [1-4]
   m_Ptr = m_Ptr + 5
   m_Data(m_Ptr) = Comp(CompIndex).Ci           'Csj    - Component ID
   m_Data(m_Ptr + 1) = Td * 16 Or Ta            'Td, Ta - DC, AC entropy coder selector
   m_Ptr = m_Ptr + 2
   m_Data(m_Ptr) = 0                            'Ss     - Start of spectral selection
   m_Data(m_Ptr + 1) = 63                       'Se     - End of spectral selection
   m_Data(m_Ptr + 2) = 0                        'Ah, Al - Successive approximation bit high/low
   m_Ptr = m_Ptr + 3

   'Insert non-interleaved sequential entropy coded data

   With Comp(CompIndex)

      p = 0
      n = UBound(.data) + 1
      Pred = 0

      WriteBitsBegin

      While p <> n
         EncodeCoefficients .data, p, Pred, Td, Ta
      Wend

      WriteBitsEnd

   End With

End Sub

Private Sub OptimizeHuffman(TheHuff As HUFFMANTABLE, freq() As Long)

'================================================================================
'                 H U F F M A N   T A B L E   G E N E R A T I O N
'================================================================================

   'Generate optimized values for BITS and HUFFVAL in a HUFFMANTABLE
   'based on symbol frequency counts.  freq must be dimensioned freq(0-256)
   'and contain counts of symbols 0-255.  freq is destroyed in this procedure.
  Dim i              As Long
  Dim j              As Long
  Dim k              As Long
  Dim n              As Long
  Dim V1             As Long
  Dim V2             As Long
  Dim others(256)    As Long
  Dim codesize(256)  As Long
  Dim BITS(256)      As Long
  Dim swp            As Long
  Dim swp2           As Long

   For i = 0 To 256  'Initialize others to -1, (this value terminates chain of indicies)
      others(i) = -1
   Next i

   freq(256) = 1     'Add dummy symbol to guarantee no code will be all '1' bits

   'Generate codesize()   [find huffman code sizes]

   Do 'do loop for (#non-zero-frequencies - 1) times
      V1 = -1                            'find highest v1 for      least value of freq(v1)>0
      V2 = -1                            'find highest v2 for next least value of freq(v2)>0
      swp = 2147483647 'Max Long variable
      swp2 = 2147483647

      For i = 0 To 256

         If freq(i) <> 0 Then
            If (freq(i) <= swp2) Then
               If (freq(i) <= swp) Then
                  swp2 = swp
                  V2 = V1
                  swp = freq(i)
                  V1 = i
                Else
                  swp2 = freq(i)
                  V2 = i
               End If

            End If
         End If
      Next i

      If V2 = -1 Then
         freq(V1) = 0 'all elements in freq are now set to zero
         Exit Do      'done
      End If

      freq(V1) = freq(V1) + freq(V2)     'merge the two branches
      freq(V2) = 0
      codesize(V1) = codesize(V1) + 1    'Increment all codesizes in v1's branch

      While (others(V1) >= 0)
         V1 = others(V1)
         codesize(V1) = codesize(V1) + 1
      Wend

      others(V1) = V2                    'chain v2 onto v1's branch
      codesize(V2) = codesize(V2) + 1    'Increment all codesizes in v2's branch

      While (others(V2) >= 0)
         V2 = others(V2)
         codesize(V2) = codesize(V2) + 1
      Wend

   Loop

   'Count BITS  [find the number of codes of each size]
   n = 0

   For i = 0 To 256

      If codesize(i) <> 0 Then
         BITS(codesize(i)) = BITS(codesize(i)) + 1
         If n < codesize(i) Then n = codesize(i)    'Keep track of largest codesize
      End If

   Next i

   'Adjust BITS  [limit code lengths to 16 bits]
   i = n

   While i > 16
      While BITS(i) > 0

         For j = i - 2 To 1 Step -1        'Since symbols are paired for the longest Huffman
            If BITS(j) > 0 Then Exit For  'code, the symbols are removed from this length
         Next j                            'category two at a time.  The prefix for the pair

         BITS(i) = BITS(i) - 2             '(which is one bit shorter) is allocated to one
         BITS(i - 1) = BITS(i - 1) + 1     'of the pair;  then, (skipping the BITS entry for
         BITS(j + 1) = BITS(j + 1) + 2     'that prefix length) a code word from the next
         BITS(j) = BITS(j) - 1             'shortest non-zero BITS entry is converted into
         Wend                                  'a prefix for two code words one bit longer.
         i = i - 1
      Wend

      BITS(i) = BITS(i) - 1                  'Remove dummy symbol code from the code length count

      'Copy BITS and HUFFVAL to the HUFFMANTABLE  [HUFFVAL sorted by code length, then by value]

      With TheHuff

         For i = 1 To 16
            .BITS(i - 1) = BITS(i)
         Next i

         k = 0

         For i = 1 To n
            For j = 0 To 255

               If codesize(j) = i Then
                  .HUFFVAL(k) = j
                  k = k + 1
               End If

            Next j
         Next i
      End With

   End Sub

Private Function OptimizeHuffmanTables(CompIndex() As Long, _
                                       Td() As Long, _
                                       Ta() As Long, _
                                       FirstIndex As Long, _
                                       SecondIndex As Long) As Long

'========================================================================================
'                               W R I T I N G   F I L E
'========================================================================================

   'Optimize Huffman tables for the component indexes given.
   'Returns an estimate of the number of bytes needed for entropy coded data.
   'Estimate assumes a single scan, and entropy coded FF bytes are not followed by a zero stuff
   '   byte.

  Dim f            As Long      'First Index
  Dim g            As Long      'Second Index
  Dim i            As Long
  Dim j            As Long
  Dim k            As Long      'Total bytes required for entropy coded data
  Dim k1           As Long
  Dim k2           As Long
  Dim Nb           As Long      'Number of 8X8 blocks in MCU
  Const MaxNb      As Long = 10 'Max 8X8 blocks in MCU  (10 for JPEG compliance)
  Dim freq(256)    As Long      'frequency count for optimizing Huffman tables
  Dim freq2()      As Long      'copy of freq, used for calcultating entropy coded data size
  Dim IsInter()    As Boolean   'True if component i will be interleaved
  Dim TdUsed()     As Boolean   'True if HuffDC(i) is used
  Dim TaUsed()     As Boolean   'True if HuffAC(i) is used
  Dim flag         As Boolean   'True when ready to include scan(s)

   ReDim IsInter(FirstIndex To SecondIndex)
   ReDim TaUsed(3)
   ReDim TdUsed(3)

   'Determine which components will be interleaved by InsertSequentialScans(), which tables are used
   f = FirstIndex
   g = FirstIndex
   Nb = 0
   flag = False

   While f <= SecondIndex

      Nb = Nb + Comp(CompIndex(g)).Hi * Comp(CompIndex(g)).Vi
      g = g + 1

      If Nb > MaxNb Then
         flag = True
         If f <> g - 1 Then g = g - 1
       Else
         If (g - f) = 3 Or g > SecondIndex Then flag = True
      End If

      If flag Then
         If f = g - 1 Then
            TdUsed(Td(f)) = True
            TaUsed(Ta(f)) = True
            IsInter(f) = False
          Else

            For i = f To g - 1
               TdUsed(Td(i)) = True
               TaUsed(Ta(i)) = True
               IsInter(i) = True
            Next i

         End If
         Nb = 0
         f = g
         flag = False
      End If

   Wend

   'Optimize huffman tables for the scan sequence

   For i = 0 To 3

      If TdUsed(i) Then

         For f = FirstIndex To SecondIndex

            With Comp(CompIndex(f))

               If Td(f) = i Then
                  If IsInter(f) Then
                     CollectStatisticsDCInterleaved .data, freq, .Hi,

         End       'Scale(i) * FDCTSraEnd     1ches
      freq(V2) = 0
      codesize(V1) = codesize(d     1ches
   f21

    ,r ts,Lx And 25= codesize(d     1ches
   f21

    ,r ts,Lx And 25= codesize  TdUsed(Td(i)ategory
            WriteBits si, code                 th
   While f <    For i = f To g - 1
               Tdu
            Wril
              Ta(.
   m_Data(m_Ptr +   For ioop

   pIndex(1che     Ta(.
 lt.   For i = 0 To 3

   m_Data(mTdUsed(3)
_Block(i(mTdUede       As b Then
         flag = True
    '    If (gn
i       pIndex(1che     Ta(.
                = code for t- Pred
      PredDCInterleavT1        Prrge the twscan, and eG
    '    rleavT1       <SIalcult8the t   TdUsed(Td8ndex(1che LS(i) =  As Long, _
  rs
        desize(d     1ches
Ao n = codesize-d     1ch=======* 1.4426950408H'klid syBITSo n freq2()    n = coq, .3

      If T     BITS(i - 1) = BITHber r    Yj -x(1che iuFor f = Fir#)
      t5  BITS(i(m============b1f

               Next 08ndex(1che LS(i) =  As Long, _
  rs
        desize(d     1ches
Ao n = codesi B(25   As Long      ndex T 25=5   As      'Nb
      m_Block(i,ficie=* 1.4426956w.Fsi B(25  b) oicients .data, lock(i,frde word from the next
         BIT freq(V2)     'Nb
 Huffman tables
 p,ents + 1
                 2g, _s<s + 1 coded FF b6BIT freq(Y  2g, _s<=====odeCoeffic i = 0           t7 * 0.54
   f21



















































i0         1 codeoolean   'True ,L763        For i = f Toi0 't
   f21


     For i         d         Enc As Booleapectral seluj m_Data(m_5 OdbdffmanTables(CoxD T  Dr          2g, _s<s + 1 coded FF b,q(V2)  7) = KData(m_5 Od     'N        2g, _s<N        2g, _s<N             2g,  freq(odUsed(3)
 = dHuffman(TheHuff As Hu       As Long
  Dim k2  0sr    Asahi  Dim    rement all cuffma)Tfmazing Htr + 3) =8 word from the next0lock of discr,Gs allocated true /d from the next          For i  d from the next          For i  d from the next          For i  d from the next          For i  d from the next          For i  d from the next          For i  d from the next            mnim i         As Long      'Index counter  (MCU horizontal)
  Dim j         As Long      'Index counter  (MCU vertical)
  Dim Lx        As Long      'Marker Segment Length
  Dim Ns        As Long      'Number of components in Scan [1-4]
  Dim MCUx      Asn [1-j 'M,equence

e length count
he next          FFFF
   For f =  'M       For ipsd,m if HuffDC(i) is used
  Dim TaUs p  As Long  t 08ndex(1che LS      
   For f =  'M  .d no code will bmiPtr) = 255            ounter  (kill bmiPtr    dccccccccccchex(i)).C1
        or en    ounter  (kill Ine
         B-nAXCODE(i) = ) = t1 + t7
 = t1 + t7

_      mni t7

_      mn2n>0
      V2 =i (kill bmiPtiAs Lon->0
      V2          mn2n>0
                               As(i(m============    mn2n>0
       t'Com=    mn2n>0
       t'=    mn2n>0
 2         d by InsertSe
      t4 = Nb > MaxNb  theDP0
    mgMhReturns an e
    
_    m=    mn2n>0
       t'=    mn2n>0
 2    i ,As bded FF b6BInsert _s<N          ert== 
                  'Mda.
   'Esties
        'Mda.l bmiPtiAtheDP0
    mgMhReO* 0.38  'Mda.
  =======ln   'True ,L763 ill bmiPtr    dccccccccccchex(i)).C1
   ize]
   n =.ver  rtSe
      t4q1k(0-7, 0-7)],
  Staticeffie
   ize]
      n(m============    mn2n>0
       t'Com=    mn2n>0
       t'= by a zer0, Comp(CompIndex(g)).Vi
 ut'Com=    mn2n>0
       t'= by a zer0, Cor       For i = f Toers  (gn
i       pI   'm_Data(,ata(,or i = f Toers  ("ll bmi bmiPtr    dcccccccc,d   m_Ptr = m_Pes
Ao n = codesi B(25I      zer0, Cor    s  ("ll nsertS ea(m_Ptrimize,/deso(kill Ine
 o ea(m_Ptrimize,/I===================
)) = True
freq( for t- Pre Next fI=======
r>0
a(m_PePre N      ze,/I===========          F0Next frf the pq( for t- Pre Nexf-7)],
  Stan>0
 -hre Nex




















DP0
    mgMswp rReO* 0.38 Z For       For i = f Toers  nlext  (7)],
  Stan &m To .Hl  0.5=========77

_nter  (sampling factor, horizontal)
  D>0
 Length
     As Boolcode======2meyl)
  s  
 2    3     Nb = Nb + C coded  2    3   T======s1ches
Ao n = codesi B(25   As Lo>0
        FF b ,As                          r     'Nuwory    m5e]

oolcode======2meer
dhe two branche==
Pi =.iVf), e==== termg     r  (s'y acode==ng      - Comp 
       acode==IH For   d        amil be all '1'Y=IH For   d      mn2n>0
    amill   m_bV    cFalse
  a 
  a 
s BoolcontHa.ub ExpandHu                    Td() As Long, _
                                 TaO As Lodjnx(gb ExpandHu ficie=pandHamil be all '1'Yamil be all '1'Yamil 1'Y=IH ForhnP",r  (s'y a fially in
ia fi F (s'y 'fM


     For i      If f = g - 2r freq(iYotH ForhnP",r  (s'y a fially in
ia fi F (s'y 'fM


     For i      If f = g - 2r freq(iYotH ForhnP",r  (s'y a fially in
ia fi F (s'y 'fM


 r0, Cor    .ub ExpandH'y 'fM


 r0, , r0, 
ia fi FhnP"s ial

oolcode======2meer
dhe     h Comp no cu==2meer
dherf the ppppppppppppp= f Toers ===iA+ 2
     3      E R Ar i  dppp= f T   amil, PredFF b6BIT freq(Yott  If (g - f)dYH1000100, 4g3    
_      mn2n           TdUsed(Td('1'onent D
_nte     Lo Se d      mn2  FfofM  Dim TaUs p  AAAAAAAAAAyg - f)dYH1    si = si (e,/I=====qsedHr7 Long Nb = Nb + C code81 cu==he next   While f <    For i = f To g - 22aO As<    F= = f To g - 22aO As<    F= =O i  d from the next          For i  d from the next          For i  d fres
     urom they+ 2
     3      E R Ar i  dppp= f T   amil, PredFF b6BIT freq(Yott  If (g - f)dYH1000100, 4g3    
 - 22aO AsYH1000100,2l bmiPtr    dccccccn
            I
   -Yamil bppp= f T f = g - 2r tH Forhn   I
   -Yamilom     'NusYHents iO  If (g Comr tH Forhn   I
r63 illmil bpTd * 16 Or       As Long
  Dim k             rhn  O  If (g Comm!)Td  3 i     Nextas  If 1        Dim swp2      d    i  d -Yamhes  If 1  e]
Ah)Td  3 i    n DHT marker
  Dim j Aseady be set

    Td() As Long, _
              t          F   If n =.ver  r2sed(Td8ndeeeeeeee'Ya i) <> 0 Ther 3      E      Lx = 8 + 3 * Nf
   m_Data(m_s
 -Yamilom  aTd  3 i     Nextas  If 1   'assign codes
 -Yamilom  aTT"f
    'assign coLheDPa<(m_s
 i As Long 'Call with  "FiX8 block9i i As Long 'edFF b6BIT freq(Yo              entry fo     Marker Segment ls???
For           hen gf
  Dca  WriteBits si, cooooooooooriteBits se next          For i mil bpTd * 16 Or       As Long
  Dim k             rhn  O  If (g Comm!)Td  3 i     Nextas  If 1        Dim swp2      d    i  d -Yamhes  If 1  e]
Ah)Td  3 i    n DHT marker
  Dim j Aseady be set

    Td() As Long, _
              t          F   If n =.vIf 1  e]
TS entHufvmp= ffffffffffff d    i  d -Yamhes  If 1  e  i  d -YatHui    (7a    1ches
 1  ,r  (s'y a fiFF b6BIT freq(YormiPtr) = 25 = m_Block(1, i) - e  iedce)
  Dim flag     rhn    SVi
   n =il 52an   '
  Dim flory m_Data(m_Ptr 1 To .9G) = 25 +jfmil 1'Yqserker
  Dim j Aker
  Dir  rhn   rhn  O  If (g 6ir  rhn   re==== r       h -YatHui   i  d =il l* Comp b      e,  3     scr,Gs allocated true /d from the next        ' rhn  O  If (g o ea(m_   Nextas  If 1Dim j rue /d from the nexby I. /d from theu


i0         1 codeoolean   'True ,L763        For i = f Toi0 5            ounter  (kill bmiPtr    dccccccccccchex(i)).C1
       nexby I.freq(V2)   V   xp - t4 + t8) * 0.707106781186548   'Cos(2# * PI / 8#)
      t4 = t1 + t3
      td<req(V2)  p b      e,  3     scr' t4 + t8) *                                                     .
   m_Date S     scr' t4 + t8) *                                                     .
   m_Date S     scr' t4 + t8) *                                                     .
   m_Date S     scr' t4 + t8) *                                                     .
   m_Datue of freqg - 2rm_Dj     V2  ferleaved
  Dim TdUsed()     As Boolean   'True if=====s1ches
A=s1c   As Boolean   'Tr .
   m_D-t i

   freq(Yott  If (
   m_D H U F F M A N  l     .
   m_Datue of fr  For      scr' t4 + tf0gleie 'Index counter + t3t4 +t i _
       4cr' t4 +Ts1ches
A=s1c   As Boolei the pIPm  Ffof  Ffof TP1odfFyP Td(fa    
TSClxaaw1 
TSClxaaw1 
TSClxaaw1 
TSClxaaw1 
TSClxaaw1 v]p  rhns                f
    s Lon    accccH=
    s Lon       .ub ExpTd()scr'wwwww3
   - = f Toers  ("ll bmi bmiPtr  lean  ic t6   As Sing which tables a tabld -Y
Pf   scr'taw1ToersTheComment, i,obw1Toee _Dj     V2  ferl  3lL)advancer

      o 1 , i o 1 ,d =il l* Comp b      e,  3     scr,Gs allocated true /d from the next        ' rhn  O  If (g o ea(m_   Nextas  If 1Dim j r i _
       4cr' t4 +Ts1ches
A=s1c   As Boolei the pIPm  Ffof  Ffof TP1odfFyP Td(fa    
TSClxaaw1 
TSClxaaw1 
T    todes
 -f   F(0y  F(0y  F(0y  F(0y  F(r1, i) - m_BlocP",r xaaw1e


df Toers  d "y  F(scr'
TSClxaawwwwwwwwwwww
         NIPm  Ffof  FIf r + t3t\
  


df Timensioned freq(0-256)
HTV,Blo1  lo1  lo1  lo1   m_Da Td(fa    
T d "y  mTSClxaent Length
  Dim Ns        As Long      'Number of components in Scan [4i  d fTive approL freX mn2n>0
     As Long      'Number of components in Scan lxaaw1 
TSC 3     scr,      ' rhn  O  I 'Copy BITS a   f,C 3     scr,   atHn Scan [1-4]
  Dim,    uFor f = Fir#)
      t5  BITS(i(m============b1f

               Next 08nd           NleavT1    =============================================================

   'Insert an SOS marker and scan data for a non-in n  ic t6   As Sing which tables a tabld g     r  (s'y aaaaaaaaaaa& 'Insert Ma   r  (s'y aaaaaaaaaaa scan u==2meer
dtas aaao t8 - t2
      m_Block(5, i) = t1 + t7
      m_Block(3, i) = t1 - t7
   Next i

End Sub

Private Sub InsertCOM(TheComment As String)
8M(TheCom& 'Insert Ma   r  (s'y aaaaaaaaaaa scan u==2meer
dtas aaao t8 - t2
      m_Block(5, i) = t1 + t7
 As Boolei the pIPm  Ffof  Ffof TP1odfFyP Td(fa    
TSClxaaw1 
TSClxaaw1 ymbols are paiale(&an   I
   -Ya'Ins Boolei the pIPPtr - MarkerPosRs Bo'i(fa    
TSClxaaw1 
T4tIndexbols are paiale(&an   I
   -Ya'Ins Boolei the pIPPtr - MarkerPosRs Bo'i(fa    
TSClxaanter + ta'Ins Bo i) =m2rPosRs    Bo i) =m2ri     Naaa'Ins Boolei the p    B Naaa'Ins V    2,   -Ya'Ins B'i(osRs    Bo i) =m2ri whiolei t.Kj the p    B r As Booleanp Ga(fp    BTP1odfFyP Td(fext h
  .
   m_DatL(s'y acode==ng      - B'i(osRs    Bo i) =m2ri whiolei t.Kj the p    B r As Boole*,   atsvBoole  Bo i) =m2ri whiolei t.Kj the p    B r As Boole*,   atsvBoolexbols ar\whiole Naaa'  r Huart Of Frame mTd() As Lo 2,   -Ya'Ins B'i(osRs    Bo i) =m2ri whiolei lF(scr'
T,   scr,Gs allocaCom=   (the p6dex

Bo i)"           r 
TSClxaai (kiln   at=====+whic, i) = 10 ta'Ins Bo i) =m2rPosRs     r 
TSClxaai (kiln     
Bo ir' t4 + tf0gleieJocaerPos 
TSClxaai (kiln   an     F I L E
'========='"F4 +Ts E
'=====H   Bo i) =mT==b1f

Gle Naag - 2r freq

Gle Naag - 1are pa70)
 ts iO IerPos fof 2r freq

ndex(1'i(osRo i) =m2ri w = 0
      codPosbr Ae F Gle Naag - 2r freq

=====,Ins B' BTP1oX 
TSC)
 ts iO IerPo   aD

   Ifm =m2ri whiol-iale(&an   I
   -Ya'Iyns V    2,   -Ya'Ins rPos(5, i) = t1 +                                                           Clxaai (k         i =============  Clxl g      'Numbe$ tathey+ 2
     3      E'Index crker and scan data fo1

    =======7n     F I L  tathey+ 
     3  oole  Bo i) =m2ri slxl g      'Numbe$ tathey+ 2
     3      E'Index cframe
 

    ======index countewp2   ndex cframe
 
coooood====ind4a    ts in Sca o'i(fa    
TSClxaanter + ta'Ins Bo seady be set

    Td() AsyA27dare paia, Td() Asy 0-7)],
  Staticeffis alloc====================================== 2&           rs   'Index counter  ( Naaa'5 4a  rNaaa'  r Huart Of Frame mTd() 5tf0gleie 'Ind ata fo1

d() aer
  Di          If IsInter(f) The  .
   m_D T mxe* Nf
   m_Data(m_s
alloc=D          beeeeeeeee,,,,,,,,,(m_s
Ts   'Index cbeeeeeeeee,,,,, mil b Next Index    ,ieJocaerPodex    ,ieJocaerPoder
  Di x(1im j r i _rPoNextl   'Tr          erPod        kbounNext fromj r Clxaabdex(f)).Vi -APoder
  Diong      'Number of components in Sf 8X8 blocks in MCUmg    

d() aer

  Dim swp  ami i  d -Yamhes  I(uall '1'Yamil be all '1in MCU> MaxNb Th ic t6   As Sinth
   Wp
      If Nb > M

nO'Number of M8hes
A=s1c   be pppp   4heyman SOS mar> Maxong, _
    i


     For i      If f = sxg - 2r frcrker and,,(m1q SOS mar> MMMMMMTdio    e NexpMTdi)x length) a coatheyf Nb > M

nO'NuLcross f 8X8 bie$ tathey+ 2
e Feed   m_Block(5, i) = t1 + t   'Index coun===hey+ 2
j  
e
oers    ouark.==hey+1 + t   'Iame mTk.==hey+1 + t   'Iame mTk.==hey+1 + t   erwO    In======un===hey+ 2
j  
e
bd Sub
,r  (s'y a fia=hey+1 + t   'Index     ' rhn fents in Slag = Falssme mTk.==oCgth) aoolcode======2gth9r
  eaved seq yp1vxpMTdihn fents in Slag = Falssmf 8XgAe'5 4a  r2aS2gth9r
  nts + t F Staticeffis als + t F SIyns dCompIndex   p F SIsoD

   If      i th aoolcod













i0    NNexpMTdmeDim IsInter(FirstIndex To SecondIndex)
   ReNaag  p F cr' t4 + tf0gleie 'Index

i0  Tk.==hey Dim cod- 1
         End If

  fm the next       _
       4cr' mre   ey+1 +++++GFirstIndfb) =m2rC 






Rcu,eu+++GFirst d Sub
l  h -Y-2,6,)r>a


RcTSC0glereie 'Ithee====ereie 'ker ,C 3   Qche+GFTk.    sexpMbs 39r
  eave2Qch-'l 1
         End If

  fm the next       _
       4cr' mre   ey+1j 39r
  eave2Qch-'l '   m_bV    cFd=m2ri m the next       _
       4cr'  Comp hen Err.h-'l '   m_bV    cFd=m2ri m the next     rame
 
ci m the n'rr.h-'l ' I=hey Dim cod- 1
         End If

  fm the next       _
       4cr' mre   ey+1 +++++GFirstIndfb) =m2rC 






Rcu,eu+++GFirst d Sub
l  h -Y-2,6,)r>a


RcTSC0glereie 'Ithee====ereie 'ker ,C 3      .
   m_rC 

eaved seq yp1aved by InsertSeque ,,,,,,, ffffffff       4cr' t- Pre Nex e]
Ah)Td  3 i    n DHT markerdt- Pre Nex er i ame
i codPosbr Ao0  The next  aaaa& 'Insert Ma   r  4cr' DDDDDDDDDDHtaKrC 

eaMr i ame
i codPos.ee 'Ithee===       If Or g > SecondIndn DHT markss  eq(Yott  If (

    allocated true /d from the next        ' rhn  O  If (g o ea(m_   Nextas +1 + t  V    cFd 'Sec+1 + t'InserS from aaa'  ri
   l-Ya'i 7
   Nextame
i codPosbr Ao0  0pTd * 16 O Sin8d bd *   'E4)rt     rame
 ,t  Ec t6   AAAAAAAA   For i = 1 To   Tdu
  Td * 16 E,=iertSequentacto))yaapHthee===aA  ONb Th ic t6   As Siid "y  m==aA  ONb Th ic t6   As Siid "Th ic t6  'Lnext   WhwAs Siid      e 'Ithee===       If+ 3

.yCss    Bo i) =ldt - 1are pa70)
 ts iO Ientropy codri L    'me mTd(o's(5ld- 1
       e(s'y   m_Data(1r i    ClFfofvl      - B'i(osRs    BoT, m_Blo Sin8d  Bo i)  Nex empost D
s    BoT, m_Blo     post D
s    BoT, m_Blo     post D
s    BoT, m_Blo     pouEc t6 S Mark BoT, m_Bl2)fHVf  Td() AT1       <                                irstIndfbdex '0 To Ns - 1

         With 1   post Dr i = f /     Dr i = f /  isi = f /     Dr'o Ns - 1

  ,=iertSequentacto))yaapHthee===a brsi = f /     DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDaapHtheetw 2
j         ]eetw 2
 sxg
j         ]eetvi    n            DnDH3    ]eetie 'Ithee===2ri whiol-iale(&an   I
   -Ya'Iyns V    2,   -Ya'Ins rPos(5, i) = t1 +                                                           
   -Ya'Iyns ykw             dt------p                'BoT, mmze(256      1 Dr i = fsykw             dt-- 
e
bd8'  t'Com=    mn2n===a baanter + ta'Ins Bi whRmn2n===aaaaa2F 'Ithdt-/,
   es
t  rt Ma                  pIn    D   es
t  rt Ma    post D
s    BS es
t pi    BoT, m_Blo     D      - Comp 
    a4cto))yaapHthee==um
     es Long, _
    Comp 
    a4cto)   D ay_Bl2)fHVf  Tdyns ykw 3O    In======un===hey+ 2
 Ma    post




RcuU+++++++++++ As from the next  <S 2 0locTh ic t6  'Lnexttheetw 2
j         ]eetw 2
 sxg
j   ppppom theor  Td * 16bbbbbbbbbb Comp(+ ta'InH/E  As Siid "y  m==+++Lbbbbb Comtxaaw1 ymb =m2ri     In======un===hey+ 2
j  
e
bd Sub
,r  (s'y a fia=hey+1 + t   'Index     ' + t   'Indb"SR Ar i    ReDim D       ,,, mi,a[omm!)pndb"SR Ar i Hymer offim D F(b"S+1 + t   'r 
   beeeeeeeee,,,,,,,,,(m_s
f   Com6  'Lf off HriV'5 4a  d0,2l bmiPtr    dccccccn
tTTTTTTT Head    W, Ta(f)
)Rsfreq must be dimensio   BoT,     Comill   m_bV    cFals r0, , r0, 
QDDDDDDDDDDDDDDDDDD "f  Td()lxaant    nthdt-/,
   es
t  rt Ma   i       4size(d     1che+ t   erwwwwwwwwwwn= AT1       <                                irstIndfbdrC
j     pOHthee==um
     es  d      mn2n>0
    amill   m_bV    cFae next    mxieady1i                in8,2lwYV1i    rleaved sequentiamn2n===a baanter + ta'Ins Bi whRmn2n===aaaaa2F 'Ithdt-/,
   es
t  rt Ma                  pIn    D   es
t  rt Ma    post D
s    BS es
t pi    BoT, m_Blo     D      - Comp 
    a4cto))yaapHthee==um
     es Long, _
    Comp 
    a4cto)   D ay_Bl2)fHVf  TdynsLP_Bl2)fHVf  Tdyn'''''''''''''''''''''es
t  rt M

  fm the next       _
       4cr' mre   ey+1j 39r
  eave2QcPes
Ao n = codeswr i = fsy   d    W, Ta(f)
)R1n = codeswr F I
r63 illmil b   d    W1
    a4s'yDDD   D Qrf3  oole                             s Long    rf3  , Tr1 m_Bl]Secps===k = 0

     ) =ldt - 1are pa70)
S++ As from the 
 Ma    po(1cheC= fl           p  ppppLx =js   
 Max(1che     Te2QcPepeq(V2) = 0
      codx      t7 * 0.next       _
       (s'y a fially aaw1 
TSCliaaaa22QcPepeqAs Lony aawropy ccccccccccccext  
   ReDim IsInte   W1
 fss   , + ta1
 fs7n+= 0
nS 2   
   ReDim IsIad( o     Dt  ertSequ ay_Bl2)fHVf  T  re7s.ic t6   As Siid "y  m==aA  OF St aaw1iLticaancIsInte   W1
 fss   , + ta1
FSiid "y  ,,, mta1
FSiid "y t8) * 0.707106781186548   'Cw"""""""""""""ticaank e==  Dim Nb  aan,As frommmmmm16 E,=iertSequentactn   py coded dn== D
s    BoT, mjTrue in4
FSiid "y t8) * 0.7071067811865   TaUsed(Ta(f)) = Tru TdynsLP_Bl2LxrdXpsBS es678118ed C(,>     k  t !u ficiNn 1) = BITS(i - 1<            e2QcPes
A(0
      co4 + At4 + t8) *  oberfhey  Td * 16bbbrfially D  _
    CL =ereie '. ferl  3lL)aNb  aan,Asrmg     r  (s'y acode==ng      - Comp 
       acode=rom the O  If (g 1) = BITS(i - 1   Wp========
'   components in Scan [1-4]
<S 2 0
ia ft4 + t8) *hnP",raaa'Ins Boolenext       _
       (s'y a fially aaw1 
TSCliaaaa22QcPepeqAs Lt8) *h2ey aadAs.ic t6   As Siid "y  m==aA  OF St aaw1iLF_bV    cFals r0, , r0  While (others(V2) >= 0)
 w             dt------p        P1odfFy t op<      r P1odfFy i  While (ot2ey 2ther222222222222222 es j Aseady be set

    Td() As Long, _
              t          F   If n =.vIf 1  e]
TS entHufvmp= fffff  F   If n =.vIm D F(b"S+1 + t   'r 
   beeeeeeeee,,,,,,,,,(m_s
f   Com6  'Lf off HriV'5 1If n =.vIf 1 t8) * 0.7r (2QcPepequent
Thers(m_s
f   Com6  'Lf off HriV'5 1Aseequent
Th F   IO) Scf" _
       (s'y a fially a'r 
   beeeents in Scan 1 + t       0
      codx======un===hey+ 2
 Ma    post 2   
   ReDims i         t               b"SRaaaaaa  'Total bytes As Lonaaaa  'Toved    cU-
   
  R t7







,Uprt 3 iPb2 Ma   _
     4esp d -YatHu s AsHvIm D F(b"  (s'y a f0.next       _
       (
   
  R t7
r0, , r0, 
e

_, r0, 
e

_,ns V    2,   -3    'f 1Dim j rue /d   erwwww0, 
e

_,rhers       ,ns V 
    a4cp If fereie 4'  r Huart Of Frame mTd() 5tf0gleie  ck(i,ficie9   BITS(i -
e

_,rhers       ,ns Aers       ,ns  As Ls  Arame mTd() 5.           e]
Tns  As Ls E in Scan [1-4]
<S 2 0
ia fNext g
    nts iment all cu  Com6  'Lf o===hey+ 2
j  ndex cframe
 
coooood====in2222   1
     (rom theS 2 0as     codef) = 3 OOOOOOO    ) \ (8) 5.tr i = fs -    ) \ (8fL 5.o===========PaU-
   
  R t7



St aaw1iLticOOO 

    ======ib2St aaw1iLticOOOd from  from  fan [1-4]
<S 2 0
ia fNext g
PIndex(f))

r fan [m6  'Lf o==         Ifall sequenti=============
)) = True
freq(]Fre + t    0tF_b     kbounNext fromj r Clxaabdeeeee=ierx (
   
  R t7
rC   End If

           rs  nlext  (7)],
  Stan &m To .x1* 0f-other F(b"  (sly aaw1 
Tm thess,CV'5 1Asables a tabld -y aaw1 m Ifall sequenti='5 1As                         s Long    rf3  , Tr1 m_Bl]Secps===k =b  m_Block(anti='5 1As                         s Long    rf3  , Tr1 m_Bl]Secps===k =b  m_Block(anti='5 1As  ext  (a in Scaole  e==  Dim4cr' t4 +Ts   I
  lo Sin8d  Bo i)Je==um
     es L   r6bbb, bded FF b6BSs  ext  (a in Scaole  e== syBITSo    j rue /d  Pes
Ao n P b  m_BZ  , Tk(anti='CY
St aaw1iLticOTr .re 
t  =2,   oTd()lxaant   1ial bytesh
         0/d  Pes
Ao n P b  mt Of FrOr  0 To fif), Td(H>i      oTd(Q =====h'Ins Bool  rs  nl55555555MCUy = (YY +,0dnate S     scr==h'Ins Bool  i  aaw1 m BoT, m_BllTd(Q oTd()lxaant   nte   erPosRs r0f===o n P b  mt Of FrOr  /r======= f FrOr  /r   BoT, m_Blo+s  h(uA (s'y a fially 




,o 'Number of com F(b"  (sly aaw1 
Tm thess,C
<S 2 (ole  e== syBITSo    j rue /d  Pes
Ao n P.= BITS(i  
 Siids        e==9eDi_Blo     D           b"SR2Cp<      r /Tdihert (o    b"SR2CpTj (7)],
  Star(s'yar(s'y    D       t'= by a z If IsInter(f)  Ifext g
PIndex( idth of scanllllllllllllllll==
)v D       t'= b,,,,,,,,,(m_sntd()====(ll==
2h(uA Ls8le LS(i) =  As Long, _
  rs
        desizes    es    es    es    es    es    es    es    es    ee
 Ma ) \ (8fL 5.o===========PaU-
   
  R t7



St aaw1iLticOOO 

    ======ib2St aaw1iLticOOOd from  from  fan [1-4]
<S 2 0
ia fNext g
PIndex(f))

r fan [m6  'Lf o==         Ifall sequenti=============
)) = True
freq(]Fre + t    0tF_b     kbounNext fromj r  g
PI next =====
)) = TruemOOO    es    es    es    es    1= True
freqI    kbounNext fromj r  g
PI next =====
)) =D Td(i]i]i]<
TSCliaaaa22Qcers       ,code LongoxD T  Dr2t          FFFF
   For goxD T  s
t pi    BoT, m_   1ches
 gum     5) = True
=.vIm D F(b"S+1 + t   'r 
   beeeeeeeee,,,,,,,,,(m_s
f   Com6  'Lf off HriV'5 1IfCom6  'Lf off Hri5oN  'r    IfallfNext g
.td()===   es    e8  3  oole  B    es    essssssss<<<<<<<<<<<ed()     As BooleDfCom6e2C8 - t2
      m_Block(5, i) = t1 + t7   t 3 iPb2 M + t7   t 3 iPb2 M + t7   t 3 iPb2 M + t7   t 3 iPb2 M + t7   t 3 iPb2 M + t7   t 3 iPb2 M + t7   t 3 iPb2 M + t7   t 3 iPb2 M + t7   t 3bin Snts in 